# AUTOGENERATED! DO NOT EDIT! File to edit: 01_loader.ipynb (unless otherwise specified).

__all__ = ['slugify', 'get_image_files', 'device', 'archive_loader', 'db_loader', 'treemap_loader']

# Cell
from pathlib import Path

# Cell
def slugify(filepath):
    return f'{filepath.stem}_{str(filepath.stat().st_mtime).split(".")[0]}'

def get_image_files(path):
    img_extensions = ['.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif', '.tiff', '.webp']
    return [(f, slugify(f)) for f in path.rglob('*') if f.suffix in img_extensions]

# Cell
import torch
import torchvision

# Cell
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
device

# Cell
def archive_loader(filepaths, root, device):
    dbpath = root/'memery.pt'
#     dbpath_backup = root/'memery.pt'
    db = db_loader(dbpath, device)

    current_slugs = [slug for path, slug in filepaths]
    archive_db = {i:db[item[0]] for i, item in enumerate(db.items()) if item[1]['slug'] in current_slugs}
    archive_slugs = [v['slug'] for v in archive_db.values()]
    new_files = [(str(path), slug) for path, slug in filepaths if slug not in archive_slugs]

    return(archive_db, new_files)

# Cell
def db_loader(dbpath, device):

    # check for savefile or backup and extract
    if Path(dbpath).exists():
        db = torch.load(dbpath, device)
#     elif dbpath_backup.exists():
#         db = torch.load(dbpath_backup)
    else:
        db = {}
    return(db)

# Cell
from annoy import AnnoyIndex

# Cell
def treemap_loader(treepath):
    treemap = AnnoyIndex(512, 'angular')

    if treepath.exists():
        treemap.load(str(treepath))
    else:
        treemap = None
    return(treemap)


# AUTOGENERATED! DO NOT EDIT! File to edit: 03_encoder.ipynb (unless otherwise specified).

__all__ = ['device', 'image_encoder', 'text_encoder', 'image_query_encoder']

# Cell
import torch
import clip
from tqdm import tqdm
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model, _ = clip.load("ViT-B/32", device)


# Cell
def image_encoder(img_loader, device):
    image_embeddings = torch.tensor(()).to(device)
    with torch.no_grad():
        for images, labels in tqdm(img_loader):
            batch_features = model.encode_image(images)
            image_embeddings = torch.cat((image_embeddings, batch_features)).to(device)

    image_embeddings = image_embeddings / image_embeddings.norm(dim=-1, keepdim=True)
    return(image_embeddings)

# Cell
def text_encoder(text, device):
    with torch.no_grad():
        text = clip.tokenize(text).to(device)
        text_features = model.encode_text(text)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)
    return(text_features)

# Cell
def image_query_encoder(image, device):
    with torch.no_grad():
        image_embed = model.encode_image(image.unsqueeze(0))
    image_embed = image_embed / image_embed.norm(dim=-1, keepdim=True)
    return(image_embed)


# AUTOGENERATED! DO NOT EDIT! File to edit: 06_fileutils.ipynb (unless otherwise specified).

__all__ = ['top_bottom', 'copy_topn_images']

# Cell
def top_bottom(query):
    results = predict_from_text(image_names, image_features, query)
    inv_results = sorted(results, key=lambda o: o[1])
    print(query.upper())
    n = 10
    w = 200
    print(f'top {n}')
    printi([file for file, score in results], n, w)
    print(f'bottom {n}')
    printi([file for file, score in inv_results], n, w)

# Cell
def copy_topn_images(results, outpath, n):
    for file, score in results[:n]:
        prefix = str(int(10*float(score)))
        filepath = Path(file)
        filename = '-'.join(filepath.parts[-2:])
        outfile = outpath/f'{prefix}-{filename}'
        outfile.write_bytes(filepath.read_bytes())


# AUTOGENERATED! DO NOT EDIT! File to edit: 02_crafter.ipynb (unless otherwise specified).

__all__ = ['make_dataset', 'pil_loader', 'DatasetImagePaths', 'clip_transform', 'crafter', 'preproc']

# Cell
import torch
from torchvision.datasets import VisionDataset
from PIL import Image

# Cell
def make_dataset(new_files):
    '''Returns a list of samples of a form (path_to_sample, class) and in
    this case the class is just the filename'''
    samples = []
    slugs = []
    for i, f in enumerate(new_files):
        path, slug = f
        samples.append((str(path), i))
        slugs.append((slug, i))
    return(samples, slugs)

# Cell
def pil_loader(path: str) -> Image.Image:
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert('RGB')

# Cell
class DatasetImagePaths(VisionDataset):
    def __init__(self, new_files, transforms = None):
        super(DatasetImagePaths, self).__init__(new_files, transforms=transforms)
        samples, slugs = make_dataset(new_files)
        self.samples = samples
        self.slugs = slugs
        self.loader = pil_loader
        self.root = 'file dataset'
    def __len__(self):
        return(len(self.samples))

    def __getitem__(self, index):
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transforms is not None:
            sample = self.transforms(sample)
        return sample, target

# Cell
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor, Normalize

# Cell
def clip_transform(n_px):
    return Compose([
        Resize(n_px, interpolation=Image.BICUBIC),
        CenterCrop(n_px),
        ToTensor(),
        Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
    ])

# Cell
def crafter(new_files, device, batch_size=128, num_workers=4):
    with torch.no_grad():
        imagefiles=DatasetImagePaths(new_files, clip_transform(224))
        img_loader=torch.utils.data.DataLoader(imagefiles, batch_size=batch_size, shuffle=False, num_workers=num_workers)
    return(img_loader)

# Cell
def preproc(img):
    transformed = clip_transform(224)(img)
    return(transformed)


# AUTOGENERATED! DO NOT EDIT! File to edit: 04_indexer.ipynb (unless otherwise specified).

__all__ = ['join_all', 'build_treemap', 'save_archives']

# Cell
def join_all(db, new_files, new_embeddings):
    start = len(db)
    for i, file in enumerate(new_files):
        path, slug = file
        index = i + start
        db[index] = {
            'slug': slug,
            'fpath': path,
            'embed': new_embeddings[i],
        }
    return(db)

# Cell
from annoy import AnnoyIndex

# Cell
def build_treemap(db):
    treemap = AnnoyIndex(512, 'angular')
    for k, v in db.items():
        treemap.add_item(k, v['embed'])

    # Build the treemap, with 5 trees rn
    treemap.build(5)

    return(treemap)


# Cell
import torch

# Cell
def save_archives(root, treemap, db):
    dbpath = root/'memery.pt'
    if dbpath.exists():
#         dbpath.rename(root/'memery-bak.pt')
        dbpath.unlink()
    torch.save(db, dbpath)

    treepath = root/'memery.ann'
    if treepath.exists():
#         treepath.rename(root/'memery-bak.ann')
        treepath.unlink()
    treemap.save(str(treepath))

    return(str(dbpath), str(treepath))